/*
 * sync_project_with_scm.e
 *
 * $Id: sync_project_with_scm.e#8 2008/12/23 17:59:26 chrisant $
 *
 *    Update the current project's file list, based on dir list definitions in
 *    a "ssync.ini" file in your SlickEdit config directory.  Edit the
 *    ssync.ini file to describe your projects.
 *
 *    Long term, SlickEdit needs improved native support for efficiently
 *    automatically updating the list of files in a project.
 *
 * Usage:
 *
 *    - Use "ssync" to sync the current project.
 *    - Use "ssync -a" to sync all projects in the current workspace.
 */
/*
 * Format of the SSYNC.INI file:
 *
 *    [ProjectName]
 *    workspace=WorkspaceName                (OPTIONAL LINE)
 *    dir=directory1, wildcards
 *    dir=directory2, wildcards
 *    dir=directory3, wildcards
 *    ...etc
 *    exclude=prefix1
 *    exclude=prefix2
 *    ...etc
 *
 * The "workspace=" line is optional, and can be used to disambiguate between
 * two projects with the same name that are in different workspaces.
 *
 * A "dir=" line specifies a directory to scan.  Prefix the directory with a
 * plus sign ("+") to scan recursively.  A directory can be relative (in which
 * case it is relative to the project directory) or it can be absolute.  It
 * can refer to environment variables.  It must be surrounded in double quote
 * marks if it contains commas (when in doubt, surround the directory name in
 * double quote marks).  The "wildcards" on a "dir=" line is optional and can
 * be used to filter the matching files (by default all files match).
 * Separate multiple wildcards with semicolons (for example "*.c;*.cpp;*.h").
 * Also certain special keywords in angled brackets are recognized as synonyms
 * for groups of wildcards.  See translate_wildcard_list in
 * sync_project_with_scm.e for the list of supported keywords (or to
 * add/modify the keywords).
 *
 * An "exclude=" line can specify a prefix to exclude.  The prefix is compared
 * to each file that was scanned and not filtered out some other way (hidden
 * file, wildcards, etc).  If there is a "dir=+." line, the matching files
 * will be relative to the current directory and appear as "foo\bar.h", etc.
 * You can use "exclude=fo" to exclude both "fox.c" and "foo\bar.h" (it is a
 * simple prefix string comparison).
 *
 * There can also be a special pseudo-project "[ALWAYS]" which is effectively
 * prepended at runtime to the beginning of each other project in the .ini
 * file.  This exists because mikesart wanted his SE macros and build files to
 * be in all of his projects.
 */

#include "slick.sh"

/*
 * TODO: (chrisant) The file system scan takes a long time for large projects.
 * Either make the scan an interruptible state machine on idle, or move it to
 * a DLL (possibly on a background thread) to avoid the overhead cost of
 * updating editor buffer data structures.  My old PROJ macro implemented what
 * both sync_project_with_scm.e and enhproj.e do; perhaps I should combine
 * them in SlickEdit as I did in my old editor?  This was particularly
 * convenient because pressing <F5> in the open file dialog started scanning
 * the file system on a background thread.  In SlickEdit it will be a little
 * more complicated because once the file list is complete files need to be
 * updated in the project and new files need to be tagged, but tagging can't
 * happen in the background.
 *
 * TODO: (chrisant) There are phases where this macro makes SlickEdit
 * unresponsive without giving any visual feedback to the user.  I believe the
 * user should always know why the editor is unresponsive.
 *
 * TODO: (chrisant) Measure how long it takes SlickEdit to build the hash and
 * update the project file list.  Investigate if the instant-replace logic
 * from my old PROJ macro/DLL can be used somehow in SlickEdit.
 *
 * TODO: (chrisant) Right now it loops through projects, performing all phases
 * for each project.  Consider instead looping through phases, performing each
 * phase for all projects.
 */

/**
 * Static variables.
 */
static boolean s_fReloadIni = true;
static int s_idIni = -1;
static int s_isDebugging = 0;

definit()
{
   if(arg(1) != 'L')
      s_idIni = -1;
   s_fReloadIni = true;
}

/**
 * Structure that defines one directory to scan.
 * @see translate_wildcard_list
 */
struct DIRINFO
{
   boolean fRecurse;    // Scan recursively.
   boolean fExclude;    // "dir" is really a prefix to exclude.
   _str dir;            // Directory to scan.
   _str filespec;       // Optional wildcards (see translate_wildcard_list).
};

/**
 * Structure for keeping track of cumulative totals while updating projects.
 */
struct CUMULATIVETOTALS
{
   int cAdded;
   int cRemoved;
   int cTotal;
};

/**
 * Update the file lists for the projects in the current workspace.  Only
 * projects in the "projects" array (near the top of this macro source file)
 * are updated.
 *
 * <ul>
 * <li>Use "ssync" to sync the current project.
 * <li>Use "ssync -a" to sync all projects in the current workspace.
 * </ul>
 */
_command void ssync() name_info(','VSARG2_REQUIRES_PROJECT_SUPPORT)
{
   s_fReloadIni = true;

   //$ todo: (chrisant) Parse the arguments to determine whether to enable
   // debug output.
   s_isDebugging = 0;

   if(_workspace_filename == '')
   {
      message("No workspace open.");
      return;
   }

   // Get the list of projects to sync.
   int status;
   _str arrayProjects[] = null;
   if(arg(1) == '-a')
   {
      status = _GetWorkspaceFiles(_workspace_filename,arrayProjects);
      if(status)
      {
         _message_box(nls("ssync: Unable to get projects from workspace '%s'.\n\n%s", _workspace_filename, get_message(status)));
         return;
      }
   }
   else if(arg() == 0)
   {
      if(_project_name != '')
         arrayProjects[0] = _project_name
   }
   else
   {
      _message_box("ssync: Syntax error.\n\nUsage:  ssync [-a]\n\nThe -a flag syncs all projects in the workspace.\nOtherwise only syncs the current project.");
      return;
   }
   if(!arrayProjects._length())
   {
      message("No projects to sync.");
      return;
   }

   // Loop over the projects.
   int ii;
   _str filename = '';
   _str displayname = '';
   CUMULATIVETOTALS totals;
   totals.cAdded = 0;
   totals.cRemoved = 0;
   totals.cTotal = 0;
   for (ii = 0; ii < arrayProjects._length(); ii++)
   {
      filename = _AbsoluteToWorkspace(strip_filename(arrayProjects[ii], 'E') :+ PRJ_FILE_EXT);
      displayname = GetProjectDisplayName(arrayProjects[ii]);

      message("Scanning project '"displayname"' ("ii+1"/"arrayProjects._length()")...");

      if (s_isDebugging)
         say("Scanning project '"filename"'...");

      ssync_project(filename, totals);
   }

   // Report cumulative results.
   message("Ssync completed for ":+
           arrayProjects._length() " project(s)  /  files: " :+
           totals.cAdded " added, " :+
           totals.cRemoved " removed, " :+
           totals.cTotal " total.");
}

static void ssync_project(_str project, CUMULATIVETOTALS& totals)
{
   // Add the project's dir list.
   _str wksName = strip_filename(_workspace_filename, 'PE');
   _str projName = strip_filename(project, 'PE');
   DIRINFO dir_list[];
   if(!dinfo_get_project(dir_list, projName, wksName))
   {
      _message_box(nls("ssync: No dir list for project '%s'.", projName));
      return;
   }

   // Change to the project file's directory so relative paths can work.
   int status;
   _str old_cwd = getcwd();
   _str prj_root = strip_filename(project, 'NE');
   if(prj_root != '')
   {
      status = chdir(prj_root, 1);
      if(status)
      {
         _message_box(nls("ssync: Could not change to directory '%s'.", prj_root));
         return;
      }
   }

   // Update the project file list.
   status = ssync_worker(project, dir_list, totals);
   if (status)
   {
      _message_box(nls("ssync: Error updating project '%s'.\n\n%s", project, get_message(status)));
   }

   // Restore the original working directory.
   if(prj_root != '' && old_cwd != '')
   {
      status = chdir(old_cwd, 1);
      if(status)
      {
         _message_box(nls("ssync: Could not change back to directory '%s'.", old_cwd));
         return;
      }
   }
}

/**
 * Updates the file list for the specified project.
 *
 * @param project       The project to update.
 * @param dir_list      The dir list to use to update the project file list.
 *
 * @return int          Returns status code (0 for success, else failure).
 */
static int ssync_worker(_str project, DIRINFO (&dir_list)[], CUMULATIVETOTALS& totals)
{
   // Get list of current project files.
   int projectfiles_list:[];
   ssync_getprojectfilelist(project, projectfiles_list);

   // Create our temporary view for insert_file_list().
   int filelist_view_id;
   int orig_view_id = _create_temp_view(filelist_view_id);
   fileman_mode();

   // Do an insert_file_list() for every dir_list + file_spec entry.
   //$ review: (chrisant) This can take a long time...!
   int ii;
   _str exclude_list[];
   for(ii = 0; ii < dir_list._length(); ii++)
   {
      DIRINFO dirInfo = dir_list[ii];  //$ review: (chrisant) Suspected perf degradation due to copy operation.

      // Exclude entries go in another array.
      if(dirInfo.fExclude)
      {
         exclude_list[exclude_list._length()] = stranslate(dirInfo.dir, FILESEP, FILESEP2);
         continue;
      }

      // Split out file spec list into an array.
      _str file_specs[];
      split(stranslate(dirInfo.filespec, FILESEP, FILESEP2), ';', file_specs);

      // Insert file lists.
      int jj;
      _str cmd;
      for(jj = 0; jj < file_specs._length(); jj++)
      {
         cmd = (dirInfo.fRecurse ? "+t " : "") maybe_quote_filename(dirInfo.dir FILESEP file_specs[jj]);

         if(s_isDebugging >= 2)
            say('list:    'cmd);

         insert_file_list("-v +a +p " cmd);
      }
   }

   // Kill any files from directories that match the exclude list.
   //$ PERF: (chrisant) This scans the file system, then deletes the matches.
   // This degrades performance significantly as the number of files to be
   // excluded grows.  A more efficient implementation would avoid scanning
   // directories that are going to be excluded.  Probably the "-exclude"
   // syntax in insert_file_list can help, otherwise a custom scan routine
   // will be needed (one may still be needed anyway, if the scan is to use an
   // interruptible state machine on idle).  Note that since I already have an
   // efficient algorithm available in a DLL, the only reason to try to
   // achieve it in Slick-C is so users on non-Windows platforms can benefit.
   for(ii = 0; ii < exclude_list._length(); ii++)
   {
      _str exclude;
      _str excludes[];
      split(exclude_list[ii], ";", excludes);

      int jj;
      for(jj = 0; jj < excludes._length(); jj++)
      {
         top();
         up();

         exclude = absolute(excludes[jj]);

         if(s_isDebugging >= 1)
            say('exclude: "'exclude'"');

         while(!search('^[ \t]+' _escape_re_chars(exclude), '@r' _fpos_case))
            delete_line();
      }
   }

   // Remove duplicate files.
   int file_list_hash:[];
   file_list_hash._makeempty();

   top();
   up();
   while(!down())
   {
      _str filename;

      get_line(filename);
      if(s_isDebugging >= 6)
         messageNwait('got "'filename'"');
      filename = stranslate(filename, FILESEP, FILESEP2);
      filename = strip(filename, 'B');
      if(_fpos_case != '')
      {
         filename = lowcase(filename);
      }

      if(projectfiles_list._indexin(filename))
      {
         // If this file is current in the project, set the projectfiles_list
         //  entry to 1 (so we don't remove it from the project) and then
         //  remove this entry from the temp buffer add list.
         projectfiles_list:[filename] = 1;

         delete_line();
         up();
      }
      else if(file_list_hash._indexin(filename) &&
         file_list_hash:[filename] != p_line)
      {
         // This is a new file and it's a duplicate so whack it.
         delete_line();
         up();
      }
      else
      {
         // New file: record the line number we first saw it at.
         file_list_hash:[filename] = p_line;
      }
   }

   // Create a list of files to remove from the project. Every hash entry
   //  in projectfiles_list[] that is a 0 should be removed.
   _str filename;
   _str filelist_remove = "";
   int files_removed = 0;
   for(filename._makeempty();;)
   {
      projectfiles_list._nextel(filename);
      if(filename._isempty())
         break;
      if(projectfiles_list:[filename] != 0)
         continue;

      if(s_isDebugging >= 1)
         say('remove:  "'filename'"');

      files_removed++;
      //$ PERF: (chrisant) String growth has exponential cost here unless
      // Slick-C is using geometric growth under the covers.  However, the
      // number of files removed should typically be small, so this may be
      // acceptable.
      filelist_remove = filelist_remove " " maybe_quote_filename(filename);
   }
   if(filelist_remove != "")
   {
      filelist_remove = strip(filelist_remove, 'B');

      int status = project_remove_filelist(project, filelist_remove);
      if(status != 0)
      {
         _message_box(nls("ssync: Warning: Unable to remove files from project.\n\n%s", get_message(status)));
      }
   }

   // Add the new list of files to our project.
   if(file_list_hash._length())
   {
      if(s_isDebugging >= 1)
      {
         _str line;
         filelist_view_id.top();
         do
         {
            filelist_view_id.get_line(line);
            if(line != '')
               say('add:     "'strip(line)'"');
         }
         while(!filelist_view_id.down());
      }

      tag_add_viewlist(_GetWorkspaceTagsFilename(), filelist_view_id, project);
   }
   else
   {
      _delete_temp_view(filelist_view_id);
   }

   int project_filecount = projectfiles_list._length() - files_removed + file_list_hash._length();

#if 0
   message("Ssync completed: " file_list_hash._length() " files added, " :+
           files_removed " files removed, " :+
           project_filecount " files total.");
#endif

   totals.cAdded += file_list_hash._length();
   totals.cRemoved += files_removed;
   totals.cTotal += project_filecount;

   activate_window(orig_view_id);
   return(0);
}

/**
 * Translates the specified <i>wildcard_list</i> if it is a special phrase:
 * <ul>
 *    <li>"<sources>" - Expands to common source file wildcards.
 *    <li>"<headers>" - Expands to common header file wildcards.
 *    <li>"<slick>" - Expands to Slick-C wildcards and some others (e.g. to
 *    pick up the sample files).
 * </ul>
 *
 * @param wildcard_list    If this matches one of the special phrases, it is
 *                         replaced with the translated list.  Otherwise the
 *                         parameter is not modified.
 *
 * @return boolean         Returns true if <i>wildcard_list</i> was modified.
 */
static boolean translate_wildcard_list(_str& wildcard_list)
{
   switch(lowcase(wildcard_list))
   {
   case '':
   case '<sources>':
      wildcard_list = '*.c;*.cc;*.cpp;*.cp;*.cxx;*.c++;':+
                      '*.h;*.hh;*.hpp;*.hxx;*.inl;':+
                      '*.cs;':+
                      '*.rc;*.rc2;*.pp;*.csv;*.dlg;':+
                      '*.idl;*.odl;':+
                      '*.bat;*.cmd;*.btm;':+
                      '*.pl;*.pm;':+
                      '*.asm;*.inc;':+
                      '*.bas;*.cls;':+
                      '*.rcp;':+
                      '*.lua;':+
                      '*.txt;*.htm;*.html;*.xml;*.php;':+
                      '*.def;*.ini;*.inf;':+
                      'makefile;makefile.*;*.mak;sources;sources.*;dirs;':+
                      'jamfile;jamrules;':+
                      '*.mm*;*.ver;':+
                      '';
      break;
   case '<headers>':
      wildcard_list = '*.h;*.hh;*.hpp;*.hxx;*.inl;':+
                      '*.idl;*.odl;':+
                      '*.asm;*.inc;':+
                      '*.txt;*.htm;*.html;*.xml;*.php;':+
                      '*.def;*.ini;*.inf;':+
                      'makefile;makefile.*;*.mak;sources;sources.*;dirs;':+
                      'jamfile;jamrules;':+
                      '*.mm*;*.ver;':+
                      '';
      break;
   case '<slick>':
      wildcard_list = '*.e;*.sh;':+
                      '*.c;*.cpp;*.h;':+
                      '*.java;':+
                      '*.rc;':+
                      '*.txt;*.htm;*.html;*.xml;*.php;':+
                      '*.def;*.ini;*.inf;':+
                      'makefile;makefile.*;*.mak;':+
                      '*.mm*;*.ver;':+
                      '';
      break;
   default:
      return false;
   }
   return true;
}

/**
 * Adds elements to the <i>dir_list</i> array.<br>
 * Assumes the current buffer is the ini file temp view, and that a project
 * section is marked.
 *
 * @see DIRINFO
 * @see translate_wildcard_list
 */
static void dinfo_add(DIRINFO (&dir_list)[])
{
   _begin_select();
   _begin_line();

   _str line;
   DIRINFO di;
   int cLines = count_lines_in_selection();
   while(cLines)
   {
      get_line(line);

      if(pos('[ \t]*dir[ \t]*=[ \t]*{\+|}{?@}', line, 1, 'ir') == 1)
      {
         _str tmp1 = strip(substr(line, pos('S0'), pos('0')));
         _str tmp2 = strip(substr(line, pos('S1'), pos('1')));

         di.fRecurse = (tmp1 == '+');
         di.fExclude = false;

         if(substr(tmp2, 1, 1) == '"')
         {
            if(pos('"{[~"]@}"', tmp2, 1, 'ir'))
            {
               tmp1 = substr(tmp2, pos('S0'), pos('0'));
               tmp2 = substr(tmp2, tmp1._length() + 1);
            }
            else
            {
               tmp1 = substr(tmp2, 2);
               tmp2 = '';
            }
         }
         else if(pos('^{[~,]@}', tmp2, 1, 'ir'))
         {
            tmp1 = substr(tmp2, pos('S0'), pos('0'));
            tmp2 = substr(tmp2, tmp1._length() + 1);
         }
         else
         {
            tmp1 = tmp2;
            tmp2 = '';
         }
         di.dir = expand_env_vars(strip(tmp1));
         while(last_char(di.dir) == FILESEP)
            di.dir = substr(di.dir, 1, di.dir._length() - 1);

         di.filespec = '';
         if(!di.fExclude)
         {
            if(pos(',[ \t]*{?@}$', tmp2, 1, 'ir'))
            {
               di.filespec = strip(substr(tmp2, pos('S0'), pos('0')));
               if(pos('"', di.filespec) == 1)
               {
                  di.filespec = substr(di.filespec, 2);
                  int posEndQuote = pos('"', di.filespec);
                  if(posEndQuote)
                     di.filespec = substr(di.filespec, 1, posEndQuote - 1);
               }
               translate_wildcard_list(di.filespec);
            }
         }

         if(di.dir != '')
         {
            if(s_isDebugging >= 3)
               say('dir'(di.fRecurse ? ' recurse' : '')' "'di.dir'", filespec "'di.filespec'"');

            dir_list[dir_list._length()] = di;
         }
      }
      else if(pos('[ \t]*exclude[ \t]*={?@}', line, 1, 'ir') == 1)
      {
         di.fRecurse = false;
         di.fExclude = true;
         di.dir = strip(substr(line, pos('S0'), pos('0')));
         di.filespec = '';

         if(s_isDebugging >= 3)
            say('exclude "'di.dir'"');

         dir_list[dir_list._length()] = di;
      }

      down();
      cLines--;
   }
}

/**
 * Finds and selects section for specified project.  Assumes the current
 * buffer is the ini file temp view.
 */
static boolean dinfo_find_project(_str projName, _str wksName)
{
   while(true)
   {
      // Find project.
      _begin_line();
      if(0 != search('^[ \t]*\[[ \t]*'strip(projName)'[ \t]*\][ \t]*(;?*|)$', 'ir'))
         break;
      int sectionLine = p_line;

      // Select the matching section.
      _deselect();
      if(0 != down())
         break; // Special case:  an empty project on the last line of the file is considered to not actually exist.
      _select_line();
      if(0 == search('^[ \t]*\[', 'ir'))
         up();
      else
         bottom();
      _select_line();
      if(p_line <= sectionLine)
      {
         _deselect();
         break;
      }

      // Check that the workspace matches, if specified.
      if(0 == search('^[ \t]*workspace[ \t]*=[ \t]*{?@}$', 'irm'))
      {
         if(!strieq(strip(get_match_text(0)), strip(wksName)))
         {
            _end_select();
            if(0 != down())
               break;
            continue;
         }
      }

      return true;
   }

   return false;
}

/**
 * Copies the dir list elements for <i>projName</i> to the <i>dir_list</i>
 * array.
 *
 * @param dir_list   Array to which to append.
 * @param projName   Name of project.
 * @param wksName    Name of workspace.
 *
 * @return boolean   Returns true if <i>projName</i> is recognized, otherwise
 *                   returns false.
 *
 * @see DIRINFO
 */
static boolean dinfo_get_project(DIRINFO (&dir_list)[], _str projName, _str wksName)
{
   // Load the project definitions from the ssync.ini file.
   if(s_idIni < 0 || s_fReloadIni)
   {
      // Delete our temp view of the ini file, if we have one.
      if(s_idIni >= 0)
      {
         _delete_temp_view(s_idIni);
         s_idIni = -1;
      }
      s_fReloadIni = false;

      // Find ini file.
      _str iniFile = slick_path_search1("ssync.ini");
      if(iniFile == '')
      {
         _message_box("Unable to find ssync.ini file.");
         return false;
      }

      // Create temp view and load ini file.
      int idTmp;
      int idOrig = _create_temp_view(idTmp);
      if(idOrig == '')
      {
         _message_box("Error creating temp view of ini file.");
         return false;
      }
      s_idIni = idTmp;
      get(iniFile);
      activate_window(idOrig);
   }

   // Find the project in the ini file, then add its entries.
   boolean fProjectFound = false;
   int idOrig;
   save_selection(auto ss);
   get_window_id(idOrig);
   activate_window(s_idIni);
   top();
   if(dinfo_find_project(projName, wksName))
   {
      DIRINFO diProject[];
      DIRINFO diAlways[];

      fProjectFound = true;

      // Add the dir list entries.
      dinfo_add(diProject);

      // Insert entries from ALWAYS pseudo-project if it exists.
      top();
      if(dinfo_find_project("ALWAYS", wksName))
         dinfo_add(diAlways);

      int ii;
      for(ii = 0; ii < diAlways._length(); ii++)
         dir_list[dir_list._length()] = diAlways[ii];
      for(ii = 0; ii < diProject._length(); ii++)
         dir_list[dir_list._length()] = diProject[ii];
   }

   activate_window(idOrig);
   restore_selection(ss);
   return fProjectFound;
}

/**
 * Expand env vars in <i>s</i>.  Env vars must be in the format "%envvar%"
 * (closing % is needed).
 *
 * <p><b>Implementation Notes:</b><br>
 * The current implementation will go into an infinite loop if there is a
 * cycle in the expanded env vars.  Since I've never encountered such a
 * situation in the real world, and I don't know offhand how CMD or etc
 * resolve such situation, I'm sticking my head in the sand and pretending
 * such things don't happen.  Please try not to prove me wrong.
 *
 * @param s          String in which to expand env vars.
 *
 * @return _str      String with the expanded env vars.
 */
static _str expand_env_vars( _str s )
{
   _str new_s = '';

   while ( pos( '(%([A-Za-z_0-9]+)%)', s, 1, 'U' ) )
   {
      int cchBefore = pos( 'S1' ) - 1;
      int cchAfter = length( s ) - ( pos( 'S1' ) + pos( '1' ) - 1 );
      new_s = '';
      if ( cchBefore )
         new_s = new_s :+ substr( s, 1, cchBefore );
      new_s = new_s :+ get_env( substr( s, pos( 'S2' ), pos( '2' ) ) );
      if ( cchAfter )
         new_s = new_s :+ substr( s, length( s ) - ( cchAfter - 1 ), cchAfter );
      s = new_s;
   }

   return s;
}

/**
 * Builds a hash array of the files in the project.
 */
static void ssync_getprojectfilelist(_str ProjectName, int (&projectfiles_list):[])
{
   _str filelist = "";
   int file_view_id = 0;
   int orig_view_id = p_window_id;

   GetProjectFiles(ProjectName, file_view_id);
   p_window_id = file_view_id;

   top();
   up();

   while(!down())
   {
      _str filename;

      get_line(filename);
      filename = strip(filename, 'B');

      if(filename != "")
      {
         if(_fpos_case != '')
            filename = lowcase(filename);
         projectfiles_list:[filename] = 0;
      }
   }

   p_window_id = orig_view_id;
   _delete_temp_view(file_view_id);
}

